#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Sep 14 21:19:32 2020

@author: carlosesteveyague
"""

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import time

#define the shape of the environment (i.e., the number states)
environment_rows = 40
environment_columns = 40

#Create the Q-function (we set it to 0 for every state and action).
#It is a 3D numpy array. 2D for the state and 1D for the action.
q_values = np.zeros((environment_rows, environment_columns, 4))

#define actions
#numeric action codes: 0 = up, 1 = right, 2 = down, 3 = left
actions = ['up', 'right', 'down', 'left']

#Create a 2D numpy array to hold the rewards for each state. 
#The array contains 40 rows and 40 columns (to match the shape of the environment), and each value is initialized to -1.
rewards = np.full((environment_rows, environment_columns), -1.)

#Create the walls. Set the reward to -100 at these points
for j in range(0,2):
    rewards[j,[j for j in range(environment_columns)]] = -100
rewards[environment_rows-1,[i for i in range(environment_columns)]] = -100
rewards[[i for i in range(environment_rows)],0] = -100
rewards[[i for i in range(environment_rows)],environment_columns-1] = -100

#Set the target area, at the middle of the first row, with width 4
for j in range (0,2):
    rewards[j, [i for i in range(environment_columns//2-1, environment_columns//2+3)]] = 100. #set the reward for the target area to 100


#Create obstacles
#store the location of the obstacle in a dictionary.
#each entry of the dictionary corresponds to the locations in each row

#UNCOMMENT the following lines to ADD OBSTACLES
#obstacles = {}
#for i in range(environment_rows):
#    obstacles[i] = []
#    
#for i in range(10,18):
#    obstacles[i] = [j for j in range(6,33)]
#    
#for i in range(22,30):
#    obstacles[i] = [j for j in range(6,17)]
#    for j in range(22,33):
#        obstacles[i].append(j)
#
##Set the reward to -100 for all the points in obstacles
#for i in range(environment_rows):
#    rewards[i, [j for j in obstacles[i]]] = -100


#Define the wind area and the velocity of the wind.
#The wind area is defined between two rows
#The direction is always to the right
first_row_wind = 18
last_row_wind = 22
wind_velocity = 3 #squares per unit of time
wind = [first_row_wind, last_row_wind, wind_velocity]


#define a function that determines if the specified location is a terminal state
def is_terminal_state(current_row_index, current_column_index):
  #if the reward for this location is -1, then it is not a terminal state
  if rewards[current_row_index, current_column_index] == -1.:
    return False
  else:
    return True

#define a function that will choose a random, non-terminal starting state
def get_starting_location():
  #get a random row and column index
  current_row_index = np.random.randint(environment_rows)
  current_column_index = np.random.randint(environment_columns)
  #continue choosing random row and column indexes until a non-terminal state is identified
  while is_terminal_state(current_row_index, current_column_index):
    current_row_index = np.random.randint(environment_rows)
    current_column_index = np.random.randint(environment_columns)
  return current_row_index, current_column_index

#define an epsilon greedy algorithm that will choose which action to take next (i.e., where to move next)
def get_next_action(current_row_index, current_column_index, epsilon):
  #if a randomly chosen value between 0 and 1 is less than epsilon, 
  #then choose the most promising value from the Q-table for this state.
  if np.random.random() < epsilon:
    return np.argmax(q_values[current_row_index, current_column_index])
  else: #choose a random action
    return np.random.randint(4)

#define a function that will get the next location based on the chosen action
def get_next_location(current_row_index, current_column_index, action_index, wind=[0,0,0]):
  new_row_index = current_row_index
  new_column_index = current_column_index
  if current_row_index >= wind[0] and current_row_index < wind[1]:
      new_column_index += wind[2]
      current_column_index += wind[2]
  if actions[action_index] == 'up':
    new_row_index -= 1
  elif actions[action_index] == 'right':
    new_column_index += 1
  elif actions[action_index] == 'down':
    new_row_index += 1
  elif actions[action_index] == 'left':
    new_column_index -= 1
    
  if new_row_index < 0:
      new_row_index = 0
  if new_row_index > environment_rows - 1:
      new_row_index = environment_rows - 1
  if new_column_index < 0:
      new_column_index = 0
  if new_column_index > environment_columns - 1:
      new_column_index = environment_columns - 1
  return new_row_index, new_column_index

#Define a function that will get the shortest path between any location within the domain and the target.
#Set a maximum number of steps in case the trajectory enters in a loop. 
T = 100
#For setting the area and the velocity of the wind, see line 64 in this script
def get_shortest_path(start_row_index, start_column_index, wind=[0,0,0]):
  #return immediately if this is an invalid starting location
    if is_terminal_state(start_row_index, start_column_index):
        return []
    else: #if this is a 'legal' starting location
        t=0
        current_row_index, current_column_index = start_row_index, start_column_index
        shortest_path = []
        shortest_path.append([current_row_index, current_column_index])
        #continue moving along the path until we reach the goal
        while not is_terminal_state(current_row_index, current_column_index) and t<T:
            #get the best action to take
            action_index = get_next_action(current_row_index, current_column_index, 1.)
            #move to the next location on the path, and add the new location to the list
            current_row_index, current_column_index = get_next_location(current_row_index, current_column_index, action_index, wind)
            shortest_path.append([current_row_index, current_column_index])
            t += 1
        return shortest_path

#define training parameters
epsilon = 0.9 #the percentage of time when we should take the best action (instead of a random action)
discount_factor = 0.9 #discount factor for future rewards
learning_rate = 0.9 #the rate at which the AI agent should learn
episodes = 1000 #number of experiments during the training
#For setting the area and the velocity of the wind, see line 64 in this script
def training(episodes, wind = [0,0,0]):

    #run through 1000 training episodes
    for episode in range(episodes):
        #get the starting location for this episode
        row_index, column_index = get_starting_location()
        
        #continue taking actions (i.e., moving) until we reach a terminal state
        #(i.e., until we reach the target area or crash into a wall)
        while not is_terminal_state(row_index, column_index):
            #choose which action to take (i.e., where to move next)
            action_index = get_next_action(row_index, column_index, epsilon)
            
            #perform the chosen action, and transition to the next state (i.e., move to the next location)
            old_row_index, old_column_index = row_index, column_index #store the old row and column indexes
            row_index, column_index = get_next_location(row_index, column_index, action_index, wind)
            
            #receive the reward for moving to the new state, and calculate the temporal difference
            reward = rewards[row_index, column_index]
            old_q_value = q_values[old_row_index, old_column_index, action_index]
            temporal_difference = reward + (discount_factor * np.max(q_values[row_index, column_index])) - old_q_value
            
            #update the Q-value for the previous state and action pair
            new_q_value = old_q_value + (learning_rate * temporal_difference)
            q_values[old_row_index, old_column_index, action_index] = new_q_value
            
    print('Training complete!')
    print('Wind =', wind)


#Produces a certain number of experiments (100 by default) with the current
#approximation of Q and counts how many of them have been succesful
#For setting the area and the velocity of the wind, see line 64 in this script
def simulation(experiments = 100, wind=[0,0,0]):
    counter = 0
    for i in range(experiments):
        start_row_index, start_column_index = get_starting_location()
        path = get_shortest_path(start_row_index, start_column_index,wind)
        terminal_row_index, terminal_column_index = path[-1]
        if rewards[terminal_row_index, terminal_column_index] > 0:
            counter += 1
    return counter
    

#Creates a video of a trajectory
def simulation_plot(wind=[0,0,0]):
    image = np.copy(rewards)
    for j in range(wind[0],wind[1]):
        for i in range(0,environment_columns):
            if rewards[j,i] == -1:
                image[j, i ]=-20
                
    plt.matshow(image, cmap=matplotlib.cm.gist_ncar)
    plt.xticks(range(0))
    plt.yticks(range(0))
    plt.show()
    time.sleep(1)
    start_row_index, start_column_index = get_starting_location()
    path = get_shortest_path(start_row_index, start_column_index,wind)
    for i in range(len(path)):
        image_simu = np.copy(image)
        pos_raw, pos_col = path[i]
        image_simu[pos_raw, pos_col] = 40
        
        plt.matshow(image_simu, cmap=matplotlib.cm.gist_ncar)
        plt.xticks(range(0))
        plt.yticks(range(0))
        plt.show()
        time.sleep(0.2)
        
#Creates a plot with the percentage of successful experiments after every
#100 Q-learning iterations
def evolution_plot(wind=[0,0,0]):
    results = [simulation(1000,wind)/1000]
    for i in range(1,101):
        training(100,wind)
        results.append(simulation(1000,wind)/1000)
    X = [100*i for i in range(101)]
    plt.plot(X,results,'b-', X, np.ones(101),':k')
    plt.ylabel('Success rate')
    plt.xlabel('Number of training experiments')
    plt.axis([0, 10000, 0, 1.1])


#Train the Q-function 5000 times and create a simulation
wind = [5,15,3]
training(5000, wind)
simulation_plot(wind)

